# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_rand_imgs": "00_prep.ipynb",
         "vid_from_frames": "00_prep.ipynb",
         "Video": "00_prep.ipynb",
         "VideoDataset": "00_prep.ipynb",
         "get_rico_imgs": "00_prep.ipynb",
         "read_video_data": "00_prep.ipynb",
         "get_non_duplicate_corpus": "00_prep.ipynb",
         "generate_setting2": "00_prep.ipynb",
         "get_all_texts": "00_prep.ipynb",
         "Extractor": "01_features.ipynb",
         "SIFTExtractor": "01_features.ipynb",
         "CNNExtractor": "01_features.ipynb",
         "imagenet_normalize_transform": "03_model.ipynb",
         "get_transforms": "01_features.ipynb",
         "SimCLRExtractor": "01_features.ipynb",
         "gen_vcodebook": "01_features.ipynb",
         "gen_codebooks": "01_features.ipynb",
         "get_df": "01_features.ipynb",
         "get_bovw": "01_features.ipynb",
         "extract_features": "01_features.ipynb",
         "new_get_bovw": "01_features.ipynb",
         "calc_tf_idf": "02_eval.ipynb",
         "cosine_similarity": "02_eval.ipynb",
         "hit_rate_at_k": "02_eval.ipynb",
         "mean_reciprocal_rank": "02_eval.ipynb",
         "r_precision": "02_eval.ipynb",
         "precision_at_k": "02_eval.ipynb",
         "average_precision": "02_eval.ipynb",
         "mean_average_precision": "02_eval.ipynb",
         "recall_at_k": "02_eval.ipynb",
         "rank_stats": "02_eval.ipynb",
         "evaluate": "02_eval.ipynb",
         "get_eval_results": "02_eval.ipynb",
         "evaluate_ranking": "02_eval.ipynb",
         "sift_frame_sim": "03_model.ipynb",
         "simclr_frame_sim": "03_model.ipynb",
         "SimCLRDataset": "03_model.ipynb",
         "get_train_transforms": "03_model.ipynb",
         "get_val_transforms": "03_model.ipynb",
         "NTXEntCriterion": "03_model.ipynb",
         "SimCLRModel": "03_model.ipynb",
         "flatten_dict": "04_approach.ipynb",
         "gen_extracted_features": "04_approach.ipynb",
         "gen_tfidfs": "04_approach.ipynb",
         "gen_bovw_similarity": "04_approach.ipynb",
         "fuzzy_LCS": "04_approach.ipynb",
         "gen_lcs_similarity": "04_approach.ipynb",
         "fix_sims": "04_approach.ipynb",
         "sort_rankings": "04_approach.ipynb",
         "approach": "04_approach.ipynb",
         "compute_sims": "04_approach.ipynb",
         "logger": "05_cli.ipynb",
         "URLs": "05_cli.ipynb",
         "download": "05_cli.ipynb",
         "VWORDS": "05_cli.ipynb",
         "N_IMGS": "05_cli.ipynb",
         "N_FRAMES_TO_KEEP": "05_cli.ipynb",
         "FPS": "05_cli.ipynb",
         "BEST_DL_MODELS": "05_cli.ipynb",
         "BEST_IR_MODELS": "05_cli.ipynb",
         "BEST_MODEL_CONFIGS": "05_cli.ipynb",
         "reproduce": "05_cli.ipynb",
         "tango": "05_cli.ipynb",
         "write_json_line_by_line": "07_utils.ipynb",
         "read_csv_to_dic_list": "07_utils.ipynb",
         "read_json": "07_utils.ipynb",
         "read_json_line_by_line": "07_utils.ipynb",
         "find_file": "07_utils.ipynb",
         "write_csv_from_json_list": "07_utils.ipynb",
         "group_dict": "07_utils.ipynb",
         "load_settings": "07_utils.ipynb",
         "get_grayscale": "07_utils.ipynb",
         "remove_noise": "07_utils.ipynb",
         "thresholding": "07_utils.ipynb",
         "thresholding_med": "07_utils.ipynb",
         "dilate": "07_utils.ipynb",
         "erode": "07_utils.ipynb",
         "opening": "07_utils.ipynb",
         "canny": "07_utils.ipynb",
         "deskew": "07_utils.ipynb",
         "match_template": "07_utils.ipynb",
         "extract_text": "07_utils.ipynb",
         "preprocess_img": "07_utils.ipynb",
         "extract_frames": "07_utils.ipynb",
         "process_frame": "07_utils.ipynb",
         "execute_retrieval_run": "08_combo.ipynb",
         "run_settings": "08_combo.ipynb",
         "write_results": "08_combo.ipynb",
         "write_rankings": "08_combo.ipynb",
         "convert_results_format": "08_combo.ipynb",
         "get_info_to_ranking_results": "08_combo.ipynb",
         "tango_combined": "08_combo.ipynb"}

modules = ["prep.py",
           "features.py",
           "eval.py",
           "model.py",
           "approach.py",
           "cli.py",
           "utils.py",
           "combo.py"]

doc_url = "https://ncoop57.github.io/tango/"

git_url = "https://github.com/ncoop57/tango/tree/main/"

def custom_doc_links(name): return None
